#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Geoid Module

In Development
"""

import numpy as np
from geodepy.convert import hp2dec


# Define test grid points
nvals = np.array([(hp2dec(-31.51), hp2dec(133.48), -8.806),
                  (hp2dec(-31.51), hp2dec(133.49), -8.743),
                  (hp2dec(-31.52), hp2dec(133.48), -8.870),
                  (hp2dec(-31.52), hp2dec(133.49), -8.805)])

lat = hp2dec(-31.515996736)
long = hp2dec(133.483540489)





#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Constants Module
"""

from math import sqrt
from datetime import date

c_vac = 299792.458
k_0 = 0.9996


# Ellipsoid Constants
class Ellipsoid(object):
    def __init__(self, semimaj, inversef):
        self.semimaj = semimaj
        self.inversef = inversef
        self.f = 1 / self.inversef
        self.semimin = float(self.semimaj * (1 - self.f))
        self.ecc1sq = float(self.f * (2 - self.f))
        self.ecc2sq = float(self.ecc1sq / (1 - self.ecc1sq))
        self.ecc1 = sqrt(self.ecc1sq)
        self.n = float(self.f / (2 - self.f))
        self.n2 = self.n ** 2


# Geodetic Reference System 1980 (http://www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7019)
grs80 = Ellipsoid(6378137, 298.257222101)

# World Geodetic System 1984 (http://www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7030)
wgs84 = Ellipsoid(6378137, 298.257223563)

# Australian National Spheroid (http://www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7003)
ans = Ellipsoid(6378160, 298.25)

# International (Hayford) 1924 (http://www.epsg-registry.org/export.htm?gml=urn:ogc:def:ellipsoid:EPSG::7022)
intl24 = Ellipsoid(6378388, 297)


# Projections
class Projection(object):
    def __init__(self, falseeast, falsenorth, cmscale, zonewidth, initialcm):
        self.falseeast = falseeast
        self.falsenorth = falsenorth
        self.cmscale = cmscale
        self.zonewidth = zonewidth
        self.initialcm = initialcm


utm = Projection(500000, 10000000, 0.9996, 6, -177)


# Helmert 14 Parameter Transformation Parameters
class Transformation(object):
    def __init__(self, from_datum, to_datum, ref_epoch,
                 tx, ty, tz, sc, rx, ry, rz,
                 d_tx=0.0, d_ty=0.0, d_tz=0.0, d_sc=0.0, d_rx=0.0, d_ry=0.0, d_rz=0.0):
        self.from_datum = from_datum    # Datum Transforming From
        self.to_datum = to_datum        # Datum Transforming To
        self.ref_epoch = ref_epoch      # Reference Epoch (datetime.date Object)
        self.tx = tx                    # Translation in x (m)
        self.ty = ty                    # Translation in y (m)
        self.tz = tz                    # Translation in z (m)
        self.sc = sc                    # Scale Change (parts per million)
        self.rx = rx                    # Rotation about x (arcseconds)
        self.ry = ry                    # Rotation about y (arcseconds)
        self.rz = rz                    # Rotation about z (arcseconds)
        self.d_tx = d_tx                # Rate of change in Translation in x (m per year)
        self.d_ty = d_ty                # Rate of change in Translation in y (m per year)
        self.d_tz = d_tz                # Rate of change in Translation in z (m per year)
        self.d_sc = d_sc                # Rate of change in Scale Change (parts per million per year)
        self.d_rx = d_rx                # Rate of change in Rotation about x (arcseconds per year)
        self.d_ry = d_ry                # Rate of change in Rotation about y (arcseconds per year)
        self.d_rz = d_rz                # Rate of change in Rotation about z (arcseconds per year)

    def __repr__(self):
        return ('Transformation: '
                + 'From ' + repr(self.from_datum) + ' to ' + repr(self.to_datum) + '\n'
                + 'Reference Epoch: ' + repr(self.ref_epoch) + '\n'
                + '  tx: ' + repr(self.tx) + 'm + ' + repr(self.d_tx) + 'm/yr' + '\n'
                + '  ty: ' + repr(self.ty) + 'm + ' + repr(self.d_ty) + 'm/yr' + '\n'
                + '  tz: ' + repr(self.tz) + 'm + ' + repr(self.d_tz) + 'm/yr' + '\n'
                + '  sc: ' + repr(self.sc) + 'ppm + ' + repr(self.d_sc) + 'ppm/yr' + '\n'
                + '  rx: ' + repr(self.rx) + '\" + ' + repr(self.d_rx) + '\"/yr' + '\n'
                + '  ry: ' + repr(self.ry) + '\" + ' + repr(self.d_ry) + '\"/yr' + '\n'
                + '  rz: ' + repr(self.rz) + '\" + ' + repr(self.d_rz) + '\"/yr' + '\n')

    def __neg__(self):
        """
        Reverses Direction of Transformation Object
        i.e. ITRF2014 to ITRF2000 transformation becomes ITRF2000 to ITRF2014 transformation
        :return: Reversed Direction Transformation Object
        """
        return Transformation(self.to_datum,
                              self.from_datum,
                              self.ref_epoch,
                              -self.tx, -self.ty, -self.tz,
                              -self.sc,
                              -self.rx, -self.ry, -self.rz,
                              -self.d_tx, -self.d_ty, -self.d_tz,
                              -self.d_sc,
                              -self.d_rx, -self.d_ry, -self.d_rz)

    def __add__(self, other):
        """
        Change Transformation Epoch to a specified date.
        Advances all transformation parameters by their respective rates of change.
        :param other: datetime.date Object
        :return: Transformation object with parameters and ref epoch moved to specified date
        """
        if type(other) == date:
            timediff = (other - self.ref_epoch).days/365.25
            return Transformation(self.to_datum,
                                  self.from_datum,
                                  other,
                                  round(self.tx + (self.d_tx * timediff), 8),
                                  round(self.ty + (self.d_ty * timediff), 8),
                                  round(self.tz + (self.d_tz * timediff), 8),
                                  round(self.sc + (self.d_sc * timediff), 8),
                                  round(self.rx + (self.d_rx * timediff), 8),
                                  round(self.ry + (self.d_ry * timediff), 8),
                                  round(self.rz + (self.d_rz * timediff), 8),
                                  self.d_tx,
                                  self.d_ty,
                                  self.d_tz,
                                  self.d_sc,
                                  self.d_rx,
                                  self.d_ry,
                                  self.d_rz)
        else:
            ValueError('supports adding datetime.date objects only')


def iers2trans(itrf_from, itrf_to, ref_epoch, tx, ty, tz, sc, rx, ry, rz, d_tx, d_ty, d_tz, d_sc, d_rx, d_ry, d_rz):
    """
    Used to convert IERS transformation parameters into Transformation Class parameters.
    Note: All rotation and delta rotation terms have sign change applied
    :param itrf_from: ITRF Realization Transforming From
    :param itrf_to: ITRF Realization Transforming To
    :param ref_epoch: Reference Epoch (YYYY.DOY)
    :param tx: Translation in x (mm)
    :param ty: Translation in y (mm)
    :param tz: Translation in z (mm)
    :param sc: Scale Change (parts per billion)
    :param rx: Rotation about x (milliarcseconds)
    :param ry: Rotation about y (milliarcseconds)
    :param rz: Rotation about z (milliarcseconds)
    :param d_tx: Rate of change in Translation in x (mm per year)
    :param d_ty: Rate of change in Translation in y (mm per year)
    :param d_tz: Rate of change in Translation in z (mm per year)
    :param d_sc: Rate of change in Scale Change (parts per billion per year)
    :param d_rx: Rate of change in Rotation about x (milliarcseconds per year)
    :param d_ry: Rate of change in Rotation about y (milliarcseconds per year)
    :param d_rz: Rate of change in Rotation about z (milliarcseconds per year)
    :return: Transformation Object following Australian Convention
    """
    return Transformation(itrf_from, itrf_to, ref_epoch,
                          round(tx / 1000, 8), round(ty / 1000, 8), round(tz / 1000, 8),
                          round(sc / 1000, 8),
                          round(-rx / 1000, 8), round(-ry / 1000, 8), round(-rz / 1000, 8),
                          round(d_tx / 1000, 8), round(d_ty / 1000, 8), round(d_tz / 1000, 8),
                          round(d_sc / 1000, 8),
                          round(-d_rx / 1000, 8), round(-d_ry / 1000, 8), round(-d_rz / 1000, 8))

# GDA1994 to GDA2020 Transformation Parameters from GDA2020 Tech Manual v1.2

gda94to20 = Transformation('GDA1994', 'GDA2020', 0,
                           0.06155, -0.01087, -0.04019, -0.009994, -0.0394924, -0.0327221, -0.0328979)

# ITRF2014 to GDA2020 (Australian Plate Motion Model) Transformation Parameters from GDA2020 Tech Manual v1.2. The
# model was derived using 109 ARGN and AuScope GNSS CORS which were used to define the RVS.

itrf14togda20 = Transformation('ITRF2014', 'GDA2020', date(2020, 1, 1),
                                0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0.00150379, 0.00118346, 0.00120716)

# ATRF to GDA2020 (Australian Plate Motion Model) Transformation Parameters (as advised by N. Brown, Geoscience
# Australia). The model was derived using 109 ARGN and AuScope GNSS CORS which were used to define the RVS.

atrf_gda2020 = Transformation('ATRF', 'GDA2020', date(2020, 1, 1),
                              0, 0, 0, 0, 0, 0, 0,
                              0, 0, 0, 0, 0.00150379, 0.00118346, 0.00120716)

# GDA1994 to ITRF Transformation Parameters from Dawson and Woods (2010)
# AGD66 and AGD84 to GDA94 Transformation Parameters from GDA94 Tech Manual v2.4
# link: http://www.icsm.gov.au/datum/gda2020-and-gda94-technical-manuals

itrf08togda94 = Transformation('ITRF2008', 'GDA1994', date(1994, 1, 1),
                               -0.08468, -0.01942, 0.03201, 0.00971, -0.0004254, 0.0022578, 0.0024015,
                               0.00142, 0.00134, 0.00090, 0.000109, 0.0015461, 0.001820, 0.0011551)

itrf05togda94 = Transformation('ITRF2005', 'GDA1994', date(1994, 1, 1),
                               -0.07973, -0.00686, 0.03803, 0.006636, -0.0000351, 0.0021211, 0.0021411,
                               0.00225, -0.00062, -0.00056, 0.000294, 0.0014707, 0.0011443, 0.0011701)

itrf00togda94 = Transformation('ITRF2000', 'GDA1994', date(1994, 1, 1),
                               -0.04591, -0.02985, -0.02037, 0.00707, -0.0016705, 0.0004594, 0.0019356,
                               -0.00466, 0.00355, 0.01124, 0.000249, 0.0017454, 0.0014868, 0.001224)

itrf97togda94 = Transformation('ITRF1997', 'GDA1994', date(1994, 1, 1),
                               -0.01463, -0.02762, -0.02532, 0.006695, -0.0017893, -0.0006047, 0.0009962,
                               -0.00860, 0.00036, 0.01125, 0.000007, 0.0016394, 0.0015198, 0.0013801)

itrf96togda94 = Transformation('ITRF1996', 'GDA1994', date(1994, 1, 1),
                               0.02454, -0.03643, -0.06812, 0.006901, -0.0027359, -0.0020431, 0.0003731,
                               -0.02180, 0.00471, 0.02627, 0.000388, 0.0020203, 0.0021735, 0.0016290)

agd84togda94 = Transformation('AGD84', 'GDA94', 0,
                              -117.763, -51.510, 139.061, -0.191, -0.292, -0.443, -0.277)

agd66togda94 = Transformation('AGD1966', 'GDA1994', 0,
                              -117.808, -51.536, 137.784, -0.290, -0.303, -0.446, -0.234)

agd66togda94_act = Transformation('AGD66', 'GDA94', 0,
                                  -129.193, -41.212, 130.730, -2.955, -0.246, -0.374, -0.329)

agd66togda94_tas = Transformation('AGD66', 'GDA94', 0,
                                  -120.271, -64.543, 161.632, 2.499, -0.217, 0.067, 0.129)

agd66togda94_vicnsw = Transformation('AGD66', 'GDA94', 0,
                                     -119.353, -48.301, 139.484, -0.613, -0.415, -0.260, -0.437)

agd66togda94_nt = Transformation('AGD66', 'GDA94', 0,
                                 -124.133, -42.003, 137.400, -1.854, 0.008, -0.557, -0.178)


# ITRF2014 Parameters
# link: http://itrf.ign.fr/doc_ITRF/Transfo-ITRF2014_ITRFs.txt

itrf14to08 = iers2trans('ITRF2014', 'ITRF2008', date(2010, 1, 1),
                        1.6, 1.9, 2.4, -0.02, 0, 0, 0,
                        0.0, 0.0, -0.1, 0.03, 0, 0, 0)

itrf14to05 = iers2trans('ITRF2014', 'ITRF2005', date(2010, 1, 1),
                        2.6, 1.0, -2.3, 0.92, 0, 0, 0,
                        0.3, 0.0, -0.1, 0.03, 0, 0, 0)

itrf14to00 = iers2trans('ITRF2014', 'ITRF2000', date(2010, 1, 1),
                        0.7, 1.2, -26.1, 2.12, 0, 0, 0,
                        0.1, 0.1, -1.9, 0.11, 0, 0, 0)

itrf14to97 = iers2trans('ITRF2014', 'ITRF1997', date(2010, 1, 1),
                        7.4, -0.5, -62.8, 3.80, 0, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

itrf14to96 = iers2trans('ITRF2014', 'ITRF1996', date(2010, 1, 1),
                        7.4, -0.5, -62.8, 3.80, 0, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

itrf14to94 = iers2trans('ITRF2014', 'ITRF1994', date(2010, 1, 1),
                        7.4, -0.5, -62.8, 3.80, 0, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

itrf14to93 = iers2trans('ITRF2014', 'ITRF1993', date(2010, 1, 1),
                        -50.4, 3.3, -60.2, 4.29, -2.81, -3.38, 0.40,
                        -2.8, -0.1, -2.5, 0.12, -0.11, -0.19, 0.07)

itrf14to92 = iers2trans('ITRF2014', 'ITRF1992', date(2010, 1, 1),
                        15.4, 1.5, -70.8, 3.09, 0, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

itrf14to91 = iers2trans('ITRF2014', 'ITRF1991', date(2010, 1, 1),
                        27.4, 15.5, -76.8, 4.49, 0, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

itrf14to90 = iers2trans('ITRF2014', 'ITRF1990', date(2010, 1, 1),
                        25.4, 11.5, -92.8, 4.79, 0, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

itrf14to89 = iers2trans('ITRF2014', 'ITRF1989', date(2010, 1, 1),
                        30.4, 35.5, -130.8, 8.19, 0, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

itrf14to88 = iers2trans('ITRF2014', 'ITRF1988', date(2010, 1, 1),
                        25.4, -0.5, -154.8, 11.29, 0.1, 0, 0.26,
                        0.1, -0.5, -3.3, 0.12, 0, 0, 0.02)

# ITRF2008 Parameters
# link: http://itrf.ign.fr/doc_ITRF/Transfo-ITRF2008_ITRFs.txt

itrf08to05 = iers2trans('ITRF2008', 'ITRF2005', date(2000, 1, 1),
                        -2.0, -0.9, -4.7, 0.94, 0, 0, 0,
                        0.3, 0.0, 0.0, 0.0, 0, 0, 0)

itrf08to00 = iers2trans('ITRF2008', 'ITRF2000', date(2000, 1, 1),
                        -1.9, -1.7, -10.5, 1.34, 0, 0, 0,
                        0.1, 0.1, -1.8, 0.08, 0, 0, 0)

itrf08to97 = iers2trans('ITRF2008', 'ITRF1997', date(2000, 1, 1),
                        4.8, 2.6, -33.2, 2.92, 0, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

itrf08to96 = iers2trans('ITRF2008', 'ITRF1996', date(2000, 1, 1),
                        4.8, 2.6, -33.2, 2.92, 0, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

itrf08to94 = iers2trans('ITRF2008', 'ITRF1994', date(2000, 1, 1),
                        4.8, 2.6, -33.2, 2.92, 0, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

itrf08to93 = iers2trans('ITRF2008', 'ITRF1993', date(2000, 1, 1),
                        -24.0, 2.4, -38.6, 3.41, -1.71, -1.48, -0.30,
                        -2.8, -0.1, -2.4, 0.09, -0.11, -0.19, 0.07)

itrf08to92 = iers2trans('ITRF2008', 'ITRF1992', date(2000, 1, 1),
                        12.8, 4.6, -41.2, 2.21, 0, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

itrf08to91 = iers2trans('ITRF2008', 'ITRF1991', date(2000, 1, 1),
                        24.8, 18.6, -47.2, 3.61, 0, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

itrf08to90 = iers2trans('ITRF2008', 'ITRF1990', date(2000, 1, 1),
                        22.8, 14.6, -63.2, 3.91, 0, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

itrf08to89 = iers2trans('ITRF2008', 'ITRF1989', date(2000, 1, 1),
                        27.8, 38.6, -101.2, 7.31, 0, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

itrf08to88 = iers2trans('ITRF2008', 'ITRF1988', date(2000, 1, 1),
                        22.8, 2.6, -125.2, 10.41, 0.10, 0, 0.06,
                        0.1, -0.5, -3.2, 0.09, 0, 0, 0.02)

# ITRF2005 Parameters
# link: http://itrf.ensg.ign.fr/ITRF_solutions/2005/tp_05-00.php

itrf05to00 = iers2trans('ITRF2005', 'ITRF2000', date(2000, 1, 1),
                        0.1, -0.8, -5.8, 0.40, 0, 0, 0,
                        -0.2, 0.1, -1.8, 0.08, 0, 0, 0)

# ITRF2000 Parameters
# link: ftp://itrf.ensg.ign.fr/pub/itrf/ITRF.TP
# NOTE: This ref lists translations in centimetres. All other ITRF transformations are shown in millimetres.
# NOTE: All translations and rates of translation shown below have been converted to millimetres.

itrf00to97 = iers2trans('ITRF2000', 'ITRF1997', date(1997, 1, 1),
                        6.7, 6.1, -18.5, 1.55, 0, 0, 0,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)

itrf00to96 = iers2trans('ITRF2000', 'ITRF1996', date(1997, 1, 1),
                        6.7, 6.1, -18.5, 1.55, 0, 0, 0,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)

itrf00to94 = iers2trans('ITRF2000', 'ITRF1994', date(1997, 1, 1),
                        6.7, 6.1, -18.5, 1.55, 0, 0, 0,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)

itrf00to93 = iers2trans('ITRF2000', 'ITRF1993', date(1988, 1, 1),
                        12.7, 6.5, -20.9, 1.95, -0.39, 0.80, -1.14,
                        -2.9, -0.2, -0.6, 0.01, -0.11, -0.19, 0.07)

itrf00to92 = iers2trans('ITRF2000', 'ITRF1992', date(1988, 1, 1),
                        14.7, 13.5, -13.9, 0.75, 0, 0, -0.18,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)

itrf00to91 = iers2trans('ITRF2000', 'ITRF1991', date(1988, 1, 1),
                        26.7, 27.5, -19.9, 2.15, 0, 0, -0.18,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)

itrf00to90 = iers2trans('ITRF2000', 'ITRF1990', date(1988, 1, 1),
                        14.7, 13.5, -13.9, 0.75, 0, 0, -0.18,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)

itrf00to89 = iers2trans('ITRF2000', 'ITRF1989', date(1988, 1, 1),
                        29.7, 47.5, -73.9, 5.85, 0, 0, -0.18,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)

itrf00to88 = iers2trans('ITRF2000', 'ITRF1988', date(1988, 1, 1),
                        24.7, 11.5, -97.9, 8.95, 0, 0, -0.18,
                        0.0, -0.6, -1.4, 0.01, 0, 0, 0.02)



#!/usr/bin/env python3

import sys
from math import radians, sin, cos, sqrt, atan2, degrees
import numpy as np


def rotation_matrix(lat, lon):
    """Returns the 3x3 rotation matrix for a given latitude and longitude
    (given in decimal degrees)
    See Section 4.2.3 of the DynaNet User's Guide v3.3
    """
    (rlat, rlon) = (radians(lat), radians(lon))
    rot_matrix = np.array(
        [[-sin(rlon), -sin(rlat) * cos(rlon), cos(rlat) * cos(rlon)],
         [cos(rlon), -sin(rlat) * sin(rlon), cos(rlat) * sin(rlon)],
         [0.0, cos(rlat), sin(rlat)]]
    )
    return rot_matrix


def vcv_cart2local(vcv_cart, lat, lon):
    """Transform a 3x3 VCV from the Cartesian to the local reference frame. If
    only a column vector of variances is supplied (3x1) then the full VCV is
    padded out with zeros for the transformation. In these cases, only the
    column vector of variances (3x1) is returned.

    See Section 4.4.1 of the DynaNet User's Guide v3.3
    """
    if vcv_cart.shape[0] == 3:
        if vcv_cart.shape[1] == 1:
            vcv_cart = np.array([[vcv_cart[0,0], 0.0, 0.0],
                                 [0.0, vcv_cart[1,0], 0.0],
                                 [0.0, 0.0, vcv_cart[2,0]]])
        elif vcv_cart.shape[1] == 3:
            pass
        else:
            sys.exit('Matrix must be either 3x1 or 3x3')
    else:
         sys.exit('Matrix must be either 3x1 or 3x3')
    rot_matrix = rotation_matrix(lat, lon)
    vcv_local = rot_matrix.transpose() @ vcv_cart @ rot_matrix
    return vcv_local


def vcv_local2cart(vcv_local, lat, lon):
    """Transform a 3x3 VCV from the local to the Cartesian reference frame. If
    only a column vector of variances is supplied (3x1) then the full VCV is
    padded out with zeros for the transformation. In these cases, only the
    column vector of variances (3x1) is returned.

    See Section 4.4.1 of the DynaNet User's Guide v3.3
    """
    if vcv_local.shape[0] == 3:
        if vcv_local.shape[1] == 1:
            vcv_local = np.array([[vcv_local[0, 0], 0.0, 0.0],
                                 [0.0, vcv_local[1, 0], 0.0],
                                 [0.0, 0.0, vcv_local[2, 0]]])
        elif vcv_local.shape[1] == 3:
            pass
        else:
            sys.exit('Matrix must be either 3x1 or 3x3')
    else:
        sys.exit('Matrix must be either 3x1 or 3x3')
    rot_matrix = rotation_matrix(lat, lon)
    vcv_cart = rot_matrix @ vcv_local @ rot_matrix.transpose()
    return vcv_cart


def error_ellipse(vcv):
    """Calculate the semi-major axis, semi-minor axis, and the orientation of
    the error ellipse calculated from a 3x3 VCV
    See Section 7.3.3.1 of the DynaNet User's Guide v3.3
    """
    z = sqrt((vcv[0, 0] - vcv[1, 1])**2 + 4 * vcv[0, 1]**2)
    a = sqrt(0.5 * (vcv[0, 0] + vcv[1, 1] + z))
    b = sqrt(0.5 * (vcv[0, 0] + vcv[1, 1] - z))
    orientation = 90 - degrees(0.5 * atan2((2 * vcv[0, 1]),
                                           (vcv[0, 0] - vcv[1, 1])))

    return a, b, orientation


def circ_hz_pu(a, b):
    """Calculate the circularised horizontal PU form the semi-major and
    semi-minor axes
    """
    q0 = 1.960790
    q1 = 0.004071
    q2 = 0.114276
    q3 = 0.371625
    c = b / a
    k = q0 + q1 * c + q2 * c**2 + q3 * c**3
    r = a * k

    return r


def k_val95(dof):
    """
    Returns the Coverage Factor k for a given 1 sigma (68.27%) Standard
    Deviation to allow conversion to a 95% Standard Deviation. This
    uses a simplified table of k values rounded to 5 decimal places
    for Degrees of Freedom (DOF) in the range 1 to 120. For DOF above
    120, returns k value of 1.96 and for DOF below 1, returns k value
    for DOF = 1. Coverage Factor produced using following scipy stats
    code: stats.t.ppf(1-0.025,dof)
    :param dof: Degrees of Freedom (Number of Measurements Minus 1)
    :return: Coverage Factor k
    """
    if not isinstance(dof, int):
        raise TypeError('Degrees of Freedom must be Int')
    if dof < 1:
        return ttable_p95[0]
    elif dof > 120:
        return 1.96
    else:
        return ttable_p95[dof - 1]


ttable_p95 = ([12.7062, 4.30265, 3.18245, 2.77645, 2.57058, 2.44691,
              2.36462, 2.30600, 2.26216, 2.22814, 2.20099, 2.17881,
              2.16037, 2.14479, 2.13145, 2.11991, 2.10982, 2.10092,
              2.09302, 2.08596, 2.07961, 2.07387, 2.06866, 2.06390,
              2.05954, 2.05553, 2.05183, 2.04841, 2.04523, 2.04227,
              2.03951, 2.03693, 2.03452, 2.03224, 2.03011, 2.02809,
              2.02619, 2.02439, 2.02269, 2.02108, 2.01954, 2.01808,
              2.01669, 2.01537, 2.01410, 2.01290, 2.01174, 2.01063,
              2.00958, 2.00856, 2.00758, 2.00665, 2.00575, 2.00488,
              2.00404, 2.00324, 2.00247, 2.00172, 2.00100, 2.00030,
              1.99962, 1.99897, 1.99834, 1.99773, 1.99714, 1.99656,
              1.99601, 1.99547, 1.99495, 1.99444, 1.99394, 1.99346,
              1.99300, 1.99254, 1.99210, 1.99167, 1.99125, 1.99085,
              1.99045, 1.99006, 1.98969, 1.98932, 1.98896, 1.98861,
              1.98827, 1.98793, 1.98761, 1.98729, 1.98698, 1.98667,
              1.98638, 1.98609, 1.98580, 1.98552, 1.98525, 1.98498,
              1.98472, 1.98447, 1.98422, 1.98397, 1.98373, 1.98350,
              1.98326, 1.98304, 1.98282, 1.98260, 1.98238, 1.98217,
              1.98197, 1.98177, 1.98157, 1.98137, 1.98118, 1.98099,
              1.98081, 1.98063, 1.98045, 1.98027, 1.98010, 1.97993])



#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Geodesy Module
"""

import os, sys
import csv
from math import degrees, radians, sqrt, sin, cos, tan, asin, acos, atan, atan2
import numpy as np
from geodepy.convert import dec2hp, hp2dec
from geodepy.constants import grs80
from geodepy.transform import grid2geo, geo2grid
from geodepy.statistics import rotation_matrix


def enu2xyz(lat, lon, east, north, up):
    """Convert a column vector in the local reference frame to a column vector
    in the Cartesian reference frame.
    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :param east: in metres
    :param north: in metres
    :param up: in metres
    :return: x, y, z in metres
    """
    rot_matrix = rotation_matrix(lat, lon)
    enu = np.array([[east], [north], [up]])
    xyz = rot_matrix @ enu
    x = xyz[0, 0]
    y = xyz[1, 0]
    z = xyz[2, 0]
    return x, y, z


def xyz2enu(lat, lon, x, y, z):
    """Convert a column vector in the Cartesian reference frame to a column
    vector in the local reference frame.
    :param lat: latitude in decimal degrees
    :param lon: longitude in decimal degrees
    :param x: in metres
    :param y: in metres
    :param z: in metres
    :return: east, north, up in metres
    """
    rot_matrix = rotation_matrix(lat, lon)
    xyz = np.array([[x], [y], [z]])
    enu = rot_matrix.transpose() @ xyz
    east = enu[0, 0]
    north = enu[1, 0]
    up = enu[2, 0]
    return east, north, up


def vincdir(lat1, lon1, azimuth1to2, ell_dist, ellipsoid=grs80):
    """
    Vincenty's Direct Formula
    :param lat1: Latitude of Point 1 (Decimal Degrees)
    :param lon1: Longitude of Point 1 (Decimal Degrees)
    :param azimuth1to2: Azimuth from Point 1 to 2 (Decimal Degrees)
    :param ell_dist: Ellipsoidal Distance between Points 1 and 2 (m)
    :param ellipsoid: Ellipsoid Object
    :return: lat2: Latitude of Point 2 (Decimal Degrees),
             lon2: Longitude of Point 2 (Decimal Degrees),
             azimuth2to1: Azimuth from Point 2 to 1 (Decimal Degrees)

    Code review: 14-08-2018 Craig Harrison
    """

    azimuth1to2 = radians(azimuth1to2)

    # Equation numbering is from the GDA2020 Tech Manual v1.0

    # Eq. 88
    u1 = atan((1 - ellipsoid.f) * tan(radians(lat1)))

    # Eq. 89
    sigma1 = atan2(tan(u1), cos(azimuth1to2))

    # Eq. 90
    alpha = asin(cos(u1) * sin(azimuth1to2))

    # Eq. 91
    u_squared = cos(alpha)**2 \
        * (ellipsoid.semimaj**2 - ellipsoid.semimin**2) \
        / ellipsoid.semimin**2

    # Eq. 92
    a = 1 + (u_squared / 16384) \
        * (4096 + u_squared * (-768 + u_squared * (320 - 175 * u_squared)))

    # Eq. 93
    b = (u_squared / 1024) \
        * (256 + u_squared * (-128 + u_squared * (74 - 47 * u_squared)))

    # Eq. 94
    sigma = ell_dist / (ellipsoid.semimin * a)

    # Iterate until the change in sigma, delta_sigma, is insignificant (< 1e-9)
    # or after 1000 iterations have been completed
    two_sigma_m = 0
    for i in range(1000):

        # Eq. 95
        two_sigma_m = 2*sigma1 + sigma

        # Eq. 96
        delta_sigma = b * sin(sigma) * (cos(two_sigma_m) + (b/4)
                                        * (cos(sigma)
                                           * (-1 + 2 * cos(two_sigma_m)**2)
                                           - (b/6) * cos(two_sigma_m)
                                           * (-3 + 4 * sin(sigma)**2)
                                           * (-3 + 4 * cos(two_sigma_m)**2)))
        new_sigma = (ell_dist / (ellipsoid.semimin * a)) + delta_sigma
        sigma_change = new_sigma - sigma
        sigma = new_sigma

        if abs(sigma_change) < 1e-12:
            break

    # Calculate the Latitude of Point 2
    # Eq. 98
    lat2 = atan2(sin(u1)*cos(sigma) + cos(u1)*sin(sigma)*cos(azimuth1to2),
                 (1 - ellipsoid.f)
                 * sqrt(sin(alpha)**2 + (sin(u1)*sin(sigma)
                        - cos(u1)*cos(sigma)*cos(azimuth1to2))**2))
    lat2 = degrees(lat2)

    # Calculate the Longitude of Point 2
    # Eq. 99
    lon = atan2(sin(sigma)*sin(azimuth1to2),
                cos(u1)*cos(sigma) - sin(u1)*sin(sigma)*cos(azimuth1to2))

    # Eq. 100
    c = (ellipsoid.f/16)*cos(alpha)**2 \
        * (4 + ellipsoid.f*(4 - 3*cos(alpha)**2))

    # Eq. 101
    omega = lon - (1-c)*ellipsoid.f*sin(alpha) \
        * (sigma + c*sin(sigma)*(cos(two_sigma_m) + c*cos(sigma)
                                 * (-1 + 2*cos(two_sigma_m)**2)))

    # Eq. 102
    lon2 = float(lon1) + degrees(omega)

    # Calculate the Reverse Azimuth
    azimuth2to1 = degrees(atan2(sin(alpha), -sin(u1)*sin(sigma)
                          + cos(u1)*cos(sigma)*cos(azimuth1to2))) + 180

    return round(lat2, 11), round(lon2, 11), round(azimuth2to1, 9)


def vincinv(lat1, lon1, lat2, lon2, ellipsoid=grs80):
    """
    Vincenty's Inverse Formula
    :param lat1: Latitude of Point 1 (Decimal Degrees)
    :param lon1: Longitude of Point 1 (Decimal Degrees)
    :param lat2: Latitude of Point 2 (Decimal Degrees)
    :param lon2: Longitude of Point 2 (Decimal Degrees)
    :param ellipsoid: Ellipsoid Object
    :return: ell_dist: Ellipsoidal Distance between Points 1 and 2 (m),
             azimuth1to2: Azimuth from Point 1 to 2 (Decimal Degrees),
             azimuth2to1: Azimuth from Point 2 to 1 (Decimal Degrees)

    Code review: 14-08-2018 Craig Harrison
    """

    # Exit if the two input points are the same
    if lat1 == lat2 and lon1 == lon2:
        return 0, 0, 0

    # Equation numbering is from the GDA2020 Tech Manual v1.0

    # Eq. 71
    u1 = atan((1 - ellipsoid.f) * tan(radians(lat1)))

    # Eq. 72
    u2 = atan((1 - ellipsoid.f) * tan(radians(lat2)))

    # Eq. 73; initial approximation
    lon = radians(lon2 - lon1)
    omega = lon

    # Iterate until the change in lambda, lambda_sigma, is insignificant
    # (< 1e-12) or after 1000 iterations have been completed
    alpha = 0
    sigma = 0
    two_sigma_m = 0
    for i in range(1000):

        # Eq. 74
        sin_sigma = sqrt((cos(u2)*sin(lon))**2
                         + (cos(u1)*sin(u2) - sin(u1)*cos(u2)*cos(lon))**2)

        # Eq. 75
        cos_sigma = sin(u1)*sin(u2) + cos(u1)*cos(u2)*cos(lon)

        # Eq. 76
        sigma = atan2(sin_sigma, cos_sigma)

        # Eq. 77
        alpha = asin((cos(u1)*cos(u2)*sin(lon)) / sin_sigma)

        # Eq. 78
        two_sigma_m = acos(cos(sigma) - 2*sin(u1)*sin(u2) / cos(alpha)**2)

        # Eq. 79
        c = (ellipsoid.f / 16) * cos(alpha)**2 * (4 + ellipsoid.f
                                                  * (4 - 3*cos(alpha)**2))

        # Eq. 80
        new_lon = omega + (1 - c) * ellipsoid.f * sin(alpha) \
            * (sigma + c*sin(sigma) * (cos(two_sigma_m) + c*cos(sigma)
               * (-1 + 2*cos(two_sigma_m)**2)))
        delta_lon = new_lon - lon
        lon = new_lon

        if abs(delta_lon) < 1e-12:
            break

    # Eq. 81
    u_squared = cos(alpha)**2 \
        * (ellipsoid.semimaj**2 - ellipsoid.semimin**2) \
        / ellipsoid.semimin**2

    # Eq. 82
    a = 1 + (u_squared / 16384) \
        * (4096 + u_squared * (-768 + u_squared * (320 - 175 * u_squared)))

    # Eq. 83
    b = (u_squared / 1024) \
        * (256 + u_squared * (-128 + u_squared * (74 - 47 * u_squared)))

    # Eq. 84
    delta_sigma = b*sin(sigma) * (cos(two_sigma_m) + (b / 4)
                                  * (cos(sigma) * (-1 + 2*cos(two_sigma_m)**2)
                                  - (b / 6)*cos(two_sigma_m)
                                  * (-3 + 4*sin(sigma)**2)
                                  * (-3 + 4*cos(two_sigma_m)**2)))
    # Calculate the ellipsoidal distance
    # Eq. 85
    ell_dist = ellipsoid.semimin*a * (sigma - delta_sigma)

    # Calculate the azimuth from point 1 to point 2
    azimuth1to2 = degrees(atan2((cos(u2)*sin(lon)),
                                (cos(u1)*sin(u2)
                                 - sin(u1)*cos(u2)*cos(lon))))

    if azimuth1to2 < 0:
        azimuth1to2 = azimuth1to2 + 360

    # Calculate the azimuth from point 2 to point 1
    azimuth2to1 = degrees(atan2(cos(u1)*sin(lon),
                                (-sin(u1)*cos(u2)
                                 + cos(u1)*sin(u2)*cos(lon)))) + 180

    # Meridian Critical Case Tests
    #if lon1 == lon2 and lat1 > lat2:
    #    azimuth1to2 = 180
    #    azimuth2to1 = 0
    #elif lon1 == lon2 and lat1 < lat2:
    #    azimuth1to2 = 0
    #    azimuth2to1 = 180

    return round(ell_dist, 3), round(azimuth1to2, 9), round(azimuth2to1, 9)


def vincdir_utm(zone1, east1, north1, azimuth1to2, ell_dist, hemisphere1='south', ellipsoid=grs80):
    # Convert utm to geographic
    pt1 = grid2geo(zone1, east1, north1, hemisphere1)
    # Use vincdir
    lat2, lon2, azimuth2to1 = vincdir(pt1[0], pt1[1], azimuth1to2, ell_dist, ellipsoid)
    # Convert geographic to utm
    hemisphere2, zone2, east2, north2, psf2, gc2 = geo2grid(lat2, lon2)
    return hemisphere2, zone2, east2, north2, azimuth2to1


def vincinv_utm(zone1, east1, north1, zone2, east2, north2, hemisphere1='south', hemisphere2='south', ellipsoid=grs80):
    # Convert utm to geographic
    pt1 = grid2geo(zone1, east1, north1, hemisphere1, ellipsoid)
    pt2 = grid2geo(zone2, east2, north2, hemisphere2, ellipsoid)
    # Use vincinv
    return vincinv(pt1[0], pt1[1], pt2[0], pt2[1], ellipsoid)


def vincdirio():
    """
    No Input:
    Prompts the user for the name of a file in csv format. Data in the file
    must be in the form Latitude, Longitude of Point 1 in Degrees Minutes
    Seconds, Geodetic Azimuth from Point 1 to 2 in Degrees Minutes Seconds and
    Distance in metres with no header line.

    No Output:
    Uses the function vincdir to calculate for each row in the csv file the
    geographic coordinate (lat, long) of Point 2 and the Azimuth from Point 2
    to Point 1, all in Degrees Minutes Seconds. This data is written to a new
    file with the name <inputfile>_out.csv
    """
    # Enter Filename
    fn = input('Enter co-ordinate file:\n')
    # Open Filename
    csvfile = open(fn)
    csvreader = csv.reader(csvfile)
    # Create Output File
    fn_part = (os.path.splitext(fn))
    fn_out = fn_part[0] + '_out' + fn_part[1]
    outfile = open(fn_out, 'w')
    # Write Output
    outfilewriter = csv.writer(outfile)
    # outfilewriter.writerow(['Latitude2', 'Longitude2', 'azimuth2to1'])
    for row in csvreader:
        lat1 = hp2dec(float(row[0]))
        long1 = hp2dec(float(row[1]))
        azimuth1to2 = hp2dec(float(row[2]))
        ell_dist = float(row[3])
        lat2, long2, azimuth2to1 = vincdir(lat1, long1, azimuth1to2, ell_dist)
        lat2 = dec2hp(lat2)
        long2 = dec2hp(long2)
        azimuth2to1 = dec2hp(azimuth2to1)
        output = [lat2, long2, azimuth2to1]
        outfilewriter.writerow(output)
    # Close Files
    outfile.close()
    csvfile.close()


def vincinvio():
    # Enter Filename
    print('Enter co-ordinate file:')
    fn = input()
    # Open Filename
    csvfile = open(fn)
    csvreader = csv.reader(csvfile)
    # Create Output File
    fn_part = (os.path.splitext(fn))
    fn_out = fn_part[0] + '_out' + fn_part[1]
    outfile = open(fn_out, 'w')
    # Write Output
    outfilewriter = csv.writer(outfile)
    outfilewriter.writerow(['Ell_Dist', 'Azimuth1to2', 'Azimuth2to1'])
    for row in csvreader:
        lat1 = hp2dec(float(row[0]))
        long1 = hp2dec(float(row[1]))
        lat2 = hp2dec(float(row[2]))
        long2 = hp2dec(float(row[3]))
        ell_dist, azimuth1to2, azimuth2to1 = vincinv(lat1, long1, lat2, long2)
        azimuth1to2 = dec2hp(azimuth1to2)
        azimuth2to1 = dec2hp(azimuth2to1)
        output = (ell_dist, azimuth1to2, azimuth2to1)
        outfilewriter.writerow(output)
    # Close Files
    outfile.close()
    csvfile.close()




#!/usr/bin/env python3

"""
Geoscience Australia - Python Geodesy Package
Convert Module
"""

from math import modf, sin, cos, atan2, radians, degrees, sqrt


def dec2hp(dec):
    minute, second = divmod(abs(dec) * 3600, 60)
    degree, minute = divmod(minute, 60)
    hp = degree + (minute / 100) + (second / 10000)
    return hp if dec >= 0 else -hp


def hp2dec(hp):
    degmin, second = divmod(abs(hp) * 1000, 10)
    degree, minute = divmod(degmin, 100)
    dec = degree + (minute / 60) + (second / 360)
    return dec if hp >= 0 else -dec


class DMSAngle(object):
    def __init__(self, degree=0, minute=0, second=0.0):
        # Set sign of object based on sign of any variable
        if degree == 0:
            if str(degree)[0] == '-':
                self.sign = -1
            elif minute < 0:
                self.sign = -1
            elif second < 0:
                self.sign = -1
            else:
                self.sign = 1
        elif degree > 0:
            self.sign = 1
        else:  # degree < 0
            self.sign = -1
        self.degree = abs(int(degree))
        self.minute = abs(int(minute))
        self.second = abs(second)

    def __repr__(self):
        if self.sign == -1:
            signsymbol = '-'
        elif self.sign == 1:
            signsymbol = '+'
        else:
            signsymbol = 'error'
        return '{DMSAngle: ' + signsymbol + str(self.degree) + 'd ' + str(self.minute) + 'm ' + str(self.second) + 's}'

    def __add__(self, other):
        return dec2dms(self.dec() + other.dec())

    def __radd__(self, other):
        return dec2dms(other.dec() + self.dec())

    def __sub__(self, other):
        return dec2dms(self.dec() - other.dec())

    def __rsub__(self, other):
        return dec2dms(other.dec() - self.dec())

    def __mul__(self, other):
        try:
            return dec2dms(self.dec() * other)
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object and Int or Float')

    def __rmul__(self, other):
        try:
            return dec2dms(other * self.dec())
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object and Int or Float')

    def __truediv__(self, other):
        try:
            return dec2dms(self.dec() / other)
        except TypeError:
            raise TypeError('Division only defined between DMSAngle Object and Int or Float')

    def __abs__(self):
        return DMSAngle(self.degree, self.minute, self.second)

    def __neg__(self):
        if self.sign == 1:
            return DMSAngle(-self.degree, -self.minute, -self.second)
        else:  # sign == -1
            return DMSAngle(self.degree, self.minute, self.second)

    def __eq__(self, other):
        return self.dec() == other.dec()

    def __ne__(self, other):
        return self.dec() != other.dec()

    def __lt__(self, other):
        return self.dec() < other.dec()

    def __gt__(self, other):
        return self.dec() > other.dec()

    def dec(self):
        return self.sign * (self.degree + (self.minute / 60) + (self.second / 3600))

    def hp(self):
        return self.sign * (self.degree + (self.minute / 100) + (self.second / 10000))

    def ddm(self):
        if self.sign == 1:
            return DDMAngle(self.degree, self.minute + (self.second/60))
        else:  # sign == -1
            return -DDMAngle(self.degree, self.minute + (self.second/60))


class DDMAngle(object):
    def __init__(self, degree=0, minute=0.0):
        # Set sign of object based on sign of any variable
        if degree == 0:
            if str(degree)[0] == '-':
                self.sign = -1
            elif minute < 0:
                self.sign = -1
            else:
                self.sign = 1
        elif degree > 0:
            self.sign = 1
        else:  # degree < 0
            self.sign = -1
        self.degree = abs(int(degree))
        self.minute = abs(minute)

    def __repr__(self):
        if self.sign == -1:
            signsymbol = '-'
        elif self.sign == 1:
            signsymbol = '+'
        else:
            signsymbol = 'error'
        return '{DDMAngle: ' + signsymbol + str(self.degree) + 'd ' + str(self.minute) + 'm}'

    def __add__(self, other):
        return dec2ddm(self.dec() + other.dec())

    def __sub__(self, other):
        return dec2ddm(self.dec() - other.dec())

    def __mul__(self, other):
        try:
            return dec2ddm(self.dec() * other)
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object and Int or Float')

    def __rmul__(self, other):
        try:
            return dec2ddm(other * self.dec())
        except TypeError:
            raise TypeError('Multiply only defined between DMSAngle Object and Int or Float')

    def __truediv__(self, other):
        try:
            return dec2ddm(self.dec() / other)
        except TypeError:
            raise TypeError('Division only defined between DMSAngle Object and Int or Float')

    def __abs__(self):
        return DDMAngle(self.degree, self.minute)

    def __neg__(self):
        if self.sign == 1:
            return DDMAngle(-self.degree, -self.minute)
        else:  # sign == -1
            return DDMAngle(self.degree, self.minute)

    def __eq__(self, other):
        return self.dec() == other.dec()

    def __ne__(self, other):
        return self.dec() != other.dec()

    def __lt__(self, other):
        return self.dec() < other.dec()

    def __gt__(self, other):
        return self.dec() > other.dec()

    def dec(self):
        return self.sign * (self.degree + (self.minute / 60))

    def hp(self):
        minute_int, second = divmod(self.minute, 1)
        return self.sign * (self.degree + (minute_int / 100) + (second * 0.006))

    def dms(self):
        minute_int, second = divmod(self.minute, 1)
        return self.sign * DMSAngle(self.degree, minute_int, second * 60)


def dec2dms(dec):
    minute, second = divmod(abs(dec) * 3600, 60)
    degree, minute = divmod(minute, 60)
    return DMSAngle(degree, minute, second) if dec >= 0 else DMSAngle(-degree, minute, second)


def dec2ddm(dec):
    minute, second = divmod(abs(dec) * 3600, 60)
    degree, minute = divmod(minute, 60)
    minute = minute + (second / 60)
    return DDMAngle(degree, minute) if dec >= 0 else DDMAngle(-degree, minute)


def hp2dms(hp):
    degmin, second = divmod(abs(hp) * 1000, 10)
    degree, minute = divmod(degmin, 100)
    return DMSAngle(degree, minute, second * 10) if hp >= 0 else DMSAngle(-degree, minute, second * 10)


def hp2ddm(hp):
    degmin, second = divmod(abs(hp) * 1000, 10)
    degree, minute = divmod(degmin, 100)
    minute = minute + (second / 6)
    return DDMAngle(degree, minute) if hp >= 0 else DDMAngle(-degree, minute)


def polar2rect(r, theta):
    x = r * sin(radians(theta))
    y = r * cos(radians(theta))
    return x, y


def rect2polar(x, y):
    r = sqrt(x ** 2 + y ** 2)
    theta = atan2(x, y)
    if theta < 0:
        theta = degrees(theta) + 360
    else:
        theta = degrees(theta)
    return r, theta


def dd2sec(dd):
    minutes, seconds = divmod(abs(dd) * 3600, 60)
    degrees, minutes = divmod(minutes, 60)
    sec = (degrees * 3600) + (minutes * 60) + seconds
    return sec if dd >= 0 else -sec


# ----------------
# Legacy Functions
# dd2dms refactored as dec2hp
def dd2dms(dd):
    minutes, seconds = divmod(abs(dd) * 3600, 60)
    degrees, minutes = divmod(minutes, 60)
    dms = degrees + (minutes / 100) + (seconds / 10000)
    return dms if dd >= 0 else -dms


# dms2dd refactored as hp2dec
def dms2dd(dms):
    degmin, seconds = divmod(abs(dms) * 1000, 10)
    degrees, minutes = divmod(degmin, 100)
    dd = degrees + (minutes / 60) + (seconds / 360)
    return dd if dms >= 0 else -dd


def dec2sex(lon, lat):
    """Convert decimal degrees to sexagesimal format

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_sex(coord):
        """Function to create a sexagesimal-formatted coordinate as a string
        """
        coord = float(coord)
        if coord < 0:
            flag = -1
            coord = abs(coord)
        else:
            flag = 1
        min_sec, deg = modf(coord)
        deg = int(deg)
        deg *= flag  # deal with negatives
        sec, min = modf(min_sec * 60)
        min = round(min)
        sec *= 60
        sex_coord = '{} {:02d} {:05.2f}'.format(deg, min, sec)

        return sex_coord

# Convert the coordinates
    sex_lon = fmt_sex(lon)
    sex_lat = fmt_sex(lat)

    return sex_lon, sex_lat


def sex2dec(lon, lat):
    """Convert a sexagesimal coordinate to decimal degrees

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_dec(coord):
        """Function to piece together a decimal coordinate
        """
        coord = str(coord)
        if coord[:1] == '-':
            flag = -1
            coord = coord[1:]
        else:
            flag = 1
        deg, min, sec = coord.split()
        dec_coord = int(deg) + float(min) / 60 + float(sec) / 3600
        dec_coord = float('{:.6f}'.format(dec_coord))
        dec_coord *= flag  # deal with negatives

        return dec_coord

# Convert the coordinates
    dec_lon = fmt_dec(lon)
    dec_lat = fmt_dec(lat)

    return dec_lon, dec_lat


def sex2hp(lon, lat):
    """Convert a sexagesimal coordinate to HP notation

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_hp(coord):
        """Function to piece together a coordinate in HP notation
        """
        coord = str(coord)
        if coord[:1] == '-':
            flag = -1
            coord = coord[1:]
        else:
            flag = 1
        deg, min, sec = coord.split()
        sec = sec.replace('.', '')
        hp_coord = '{}.{:02d}{:04d}'.format(deg, int(min), int(sec))
        if flag == -1:
            hp_coord = '-' + hp_coord  # deal with negatives

        return hp_coord

# Convert the coordinates
    hp_lon = fmt_hp(lon)
    hp_lat = fmt_hp(lat)

    return hp_lon, hp_lat


def hp2sex(lon, lat):
    """Convert HP notation to a sexagesimal coordinate

    Longitudes go from -180 to 180
    Latitudes go from -90 to 90
    """
    def fmt_sex(coord):
        """Function to piece together a coordinate in HP notation
        """
        coord = str(coord)
        if coord[:1] == '-':
            flag = -1
            coord = coord[1:]
        else:
            flag = 1
        deg, min, sec = coord.split()
        sec = sec.replace('.', '')
        hp_coord = '{}.{:02d}{:04d}'.format(deg, int(min), int(sec))
        if flag == -1:
            hp_coord = '-' + hp_coord  # deal with negatives

        return hp_coord

# Convert the coordinates
    hp_lon = fmt_sex(lon)
    hp_lat = fmt_sex(lat)

    return hp_lon, hp_lat


def dd2dms_v(dd):
    minutes, seconds = divmod(abs(dd) * 3600, 60)
    degrees, minutes = divmod(minutes, 60)
    dms = degrees + (minutes / 100) + (seconds / 10000)
    dms[dd <= 0] = -dms[dd <= 0]
    return dms


def dms2dd_v(dms):
    degmin, seconds = divmod(abs(dms) * 1000, 10)
    degrees, minutes = divmod(degmin, 100)
    dd = degrees + (minutes / 60) + (seconds / 360)
    dd[dms <= 0] = -dd[dms <= 0]
    return dd


class DNACoord(object):
    def __init__(self, pointid, const, easting, northing, zone, lat,
                 long, ortho_ht, ell_ht, x, y, z, x_sd, y_sd, z_sd, desc):
        self.pointid = pointid
        self.const = const
        self.easting = easting
        self.northing = northing
        self.zone = zone
        self.lat = lat
        self.long = long
        self.ortho_ht = ortho_ht
        self.ell_ht = ell_ht
        self.x = x
        self.y = y
        self.z = z
        self.x_sd = x_sd
        self.y_sd = y_sd
        self.z_sd = z_sd
        self.desc = desc

    def converthptodd(self):
        self.lat = hp2dec(self.lat)
        self.long = hp2dec(self.long)


def read_dnacoord(fn):
    coord_list = []
    with open(fn, 'r') as file:
        dnadata = file.readlines()
        for line in dnadata:
            pointid = line[0:20]
            const = line[21:25]
            easting = float(line[28:40])
            northing = float(line[41:58])
            zone = int(line[60:63])
            lat = float(line[63:78])
            long = float(line[78:92])
            ortho_ht = float(line[93:103])
            ell_ht = float(line[103:114])
            x = float(line[115:129])
            y = float(line[130:144])
            z = float(line[145:159])
            x_sd = float(line[160:171])
            y_sd = float(line[172:181])
            z_sd = float(line[182:191])
            desc = line[192:-1]
            record = DNACoord(pointid.strip(), const.strip(), easting, northing, zone, lat,
                              long, ortho_ht, ell_ht, x, y, z, x_sd, y_sd, z_sd, desc.strip())
            coord_list.append(record)
    return coord_list


#___________________________________________________________________________#
# Some notes:
# Written by Jack McCubbine of Geoscience Australia, date: 08/11/2019
# This code contains functions to handle tranformations between GPS and 
# AWVS/AHD and Vice Versa 
# Gridded data used for the varisous reference surfaces are geotiff files
# These allow direct access remotely using "gdal"
#___________________________________________________________________________#
# Import dependencies 
import gdal
import numpy as np
from scipy.interpolate import griddata
import geodepy.Height_filenames as Height_filenames
#___________________________________________________________________________#
# Interpolation functions
def interp_file(Lat,Long,file):
    # Import the DOVPM file
    f = gdal.Open(file)
    # load band (akin to a variable in dataset)
    band = f.GetRasterBand(1)   
    # get the pixel width, height, etc.
    transform = f.GetGeoTransform()
    # Grid resolution (known)
    res=transform[1]
    # convert lat,lon to row,col
    column = (Long - transform[0]) / transform[1]
    row = (Lat - transform[3]) / transform[5]
    # get pixel values surrounding data point
    Surrounding_data=(band.ReadAsArray(np.floor(column-2), np.floor(row-2), 5, 5))
    # convert row,col back to north,east
    Long_c = transform[0] + np.floor(column) * res
    Lat_c = transform[3] - np.floor(row) * res
    # set up matrices for interpolation
    count=-1
    pos=np.zeros((25,2))
    Surrounding_data_v=np.zeros((25,1))
    for k in range(-2,3):
        for j in range(-2,3):
            count=count+1
            pos[count]=(Long_c+j*res,Lat_c-k*res)
            Surrounding_data_v[count]=Surrounding_data[k+2,j+2]          
    interp_val=griddata(pos,Surrounding_data_v,(Long,Lat),method='cubic')
    return interp_val

#___________________________________________________________________________#
# Functions to handle the conversions from one height to another
def GPS_to_AVWS(Lat,Long,GPS_H):
    zeta=interp_file(Lat,Long,Height_filenames.file_AVWS) # AVWS file
    zeta_std=interp_file(Lat,Long,Height_filenames.file_AVWS_STD) # AVWS STD file
    NORMAL_H=GPS_H-zeta
    return [NORMAL_H,zeta_std]

def AVWS_to_GPS(Lat,Long,AVWS_H):
    zeta=interp_file(Lat,Long,Height_filenames.file_AVWS) # AVWS file
    zeta_std=interp_file(Lat,Long,Height_filenames.file_AVWS_STD) # AVWS STD file
    GPS_H=AVWS_H+zeta
    return [GPS_H,zeta_std]

def AHD_to_AVWS(Lat,Long,AHD_H,file_AG2020):
    # Convert to GPS
    GPS_H=AHD_H+interp_file(Lat,Long,Height_filenames.file_AG2020) # AUSGEOID2020 file
    # Convert to AVWS
    Normal_H=GPS_H-interp_file(Lat,Long,Height_filenames.file_AVWS) # AVWS file
    return [Normal_H]

def GPS_to_AHD(Lat,Long,GPS_H):
    N=interp_file(Lat,Long,Height_filenames.file_AG2020) # AUSGEOID2020 file
    N_std=interp_file(Lat,Long,Height_filenames.file_AG2020_STD) # AUSGEOID2020 STD file
    AHD_H=GPS_H-N
    return [AHD_H,N_std]

def AHD_to_GPS(Lat,Long,AHD_H):
    N=interp_file(Lat,Long,Height_filenames.file_AG2020) # AUSGEOID2020 file
    N_std=interp_file(Lat,Long,Height_filenames.file_AG2020_STD) # AUSGEOID2020 STD file
    GPS_H=AHD_H+N
    return [GPS_H,N_std]

def AVWS_to_AHD(Lat,Long,Normal_H):   
    # Convert to GPS    
    GPS_H=Normal_H+interp_file(Lat,Long,Height_filenames.file_AVWS) # AVWS file
    # Convert to AHD
    AHD_H=GPS_H-interp_file(Lat,Long,Height_filenames.file_AG2020) # AUSGEOID2020 file
    return [AHD_H]

def DOV(Lat,Long):   
    # Convert to GPS    
    DOV_PM=interp_file(Lat,Long,Height_filenames.file_DOV_PM) # AVWS file
    # Convert to AHD
    DOV_PV=interp_file(Lat,Long,Height_filenames.file_DOV_PV) # AUSGEOID2020 file
    return [DOV_PM,DOV_PV]



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 29 08:56:15 2019

@author: u84157
"""
file_DOV_PV='/vsicurl/https://geoid.s3-ap-southeast-2.amazonaws.com/AVWS/DOV_PV.tif'
file_DOV_PM='/vsicurl/https://geoid.s3-ap-southeast-2.amazonaws.com/AVWS/DOV_PM.tif'
file_AG2020='/vsicurl/https://geoid.s3-ap-southeast-2.amazonaws.com/AVWS/AUSGeoid2020_RELEASEV20170908.tif'
file_AG2020_STD='/vsicurl/https://geoid.s3-ap-southeast-2.amazonaws.com/AVWS/AUSGeoid2020_RELEASEV20170908_err.tif'
file_AVWS='/vsicurl/https://geoid.s3-ap-southeast-2.amazonaws.com/AVWS/AVWS_20191107.tif'
file_AVWS_STD='/vsicurl/https://geoid.s3-ap-southeast-2.amazonaws.com/AVWS/AVWS_STD_20191107.tif'


import struct
from datetime import datetime as dt
import numpy as np
import pandas as pd

'''
Created 12/11/2018

@Author: Jaimie Dodd

This script defines classes for reading the AUSGeoid2020 binary NTv2 file, and interpolating the Ellipsoid to AHD 
Separation, Deflection in the Prime Meridian and Deflection in the Prime Vertical of a given latitude and longitude via
bilinear or bicubic interpolation (user specified input or bicubic by default).

Values have been tested against the output of DynAdjust.
'''

class NTv2ReaderBinary(object):
    def __init__(self):
        self.numOrec = 0
        self.numSrec = 0
        self.numFile = 0
        self.gsType = ''
        self.version = ''
        self.systemF = ''
        self.systemT = ''
        self.majorF = 0
        self.minorF = 0
        self.majorT = 0
        self.minorT = 0
        self.subgrids = pd.DataFrame(columns=['CONTAINS', 'SUB_NAME', 'PARENT', 'CREATED', 'UPDATED', 'S_LAT', 'N_LAT',
                                              'E_LONG', 'W_LONG', 'LAT_INC', 'LONG_INC', 'GS_COUNT',
                                              'ELLIPSOID_TO_AHD_SEPARATION', 'DEFLECTION_PRIME_MERIDIAN',
                                              'DEFLECTION_PRIME_VERTICAL'])
        self.ellps2ahdsep = 0
        self.deflectionprimemeridian = 0
        self.deflectionprimevertical = 0

    def ntv2reader(self, file, lat, lon, interpolation_method='bicubic'):
        """
        Function for reading in the binary NTv2 file and

        :param file: NTv2 file (full location as string)
        :param lat: latitude of point of interest (decimal degrees)
        :param lon: longitude of point of interest (decimal degrees)
        :param interpolation_method: 'bilinear' or 'bicubic' ('bicubic' by default)
        :return: tuple containing ellipsoid to AHD separation, deflection in prime meridian and deflection in prime
        vertical.

        """

        # If interpolation method specified is something other than bilinear or bicubic, then raise ValueError
        interpolation_options = ['bilinear', 'bicubic']
        if interpolation_method not in interpolation_options:
            raise ValueError("Invalid Interpolation method. Expected one of: %s" % interpolation_options)

        # Read the NTv2 file as binary
        f = open(file, 'rb')

        # NUM_OREC
        f.seek(8, 1)

        byte = f.read(4)
        self.numOrec = int.from_bytes(byte, byteorder='little')

        f.seek(4, 1)

        # NUM_SREC
        f.seek(8, 1)

        byte = f.read(4)
        self.numSrec = int.from_bytes(byte, byteorder='little')

        f.seek(4, 1)

        # NUM_FILE
        f.seek(8, 1)

        byte = f.read(4)
        self.numFile = int.from_bytes(byte, byteorder='little')

        f.seek(4, 1)

        # GS_TYPE
        f.seek(8, 1)

        byte = f.read(8)
        self.gsType = byte.decode('utf8').strip('\x00').strip()

        # VERSION
        f.seek(8, 1)

        byte = f.read(8)
        self.version = byte.decode('utf8').strip('\x00').strip()

        # SYSTEM_F
        f.seek(8, 1)

        byte = f.read(8)
        self.systemF = byte.decode('utf8').strip('\x00').strip()

        # SYSTEM_T
        f.seek(8, 1)

        byte = f.read(8)
        self.systemT = byte.decode('utf8').strip('\x00').strip()

        # MAJOR_F
        f.seek(8, 1)

        byte = f.read(8)
        self.majorF = struct.unpack('d', byte)[0]

        # MINOR_F
        f.seek(8, 1)

        byte = f.read(8)
        self.minorF = struct.unpack('d', byte)[0]

        # MAJOR_T
        f.seek(8, 1)

        byte = f.read(8)
        self.majorT = struct.unpack('d', byte)[0]

        # MINOR_T
        f.seek(8, 1)

        byte = f.read(8)
        self.minorT = struct.unpack('d', byte)[0]

        # Convert lat and lon to seconds
        if self.gsType == 'SECONDS':
            lat = lat * 3600
            lon = lon * -3600

        # Sub Grids
        for i in range(0, self.numFile):
            self.subgrids = self.subgrids.append(SubGrid().findsubgrid(f, lat, lon, interpolation_method),
                                                 ignore_index=True)

        # Close the NTv2 file
        f.close()

        # Filter subgrids dataframe so that only subgrids with values remain
        self.subgrids = self.subgrids[self.subgrids['CONTAINS']]

        # If more than one subgrid which contains coordinates, then filter dataframe so subgrid with smallest gridwidth
        # remains
        if len(self.subgrids) > 1:
            self.subgrids = self.subgrids.loc[[self.subgrids.loc[:, 'LAT_INC'].idxmin()]]

        # If subgrids dataframe is not empty, then return values for ellipsoid to AHD separation, deflection of prime
        # meridian and deflection of prime vertical
        if not self.subgrids.empty:
            self.ellps2ahdsep = self.subgrids.iloc[0, -3]
            self.deflectionprimemeridian = self.subgrids.iloc[0, -2]
            self.deflectionprimevertical = self.subgrids.iloc[0, -1]
            return self.ellps2ahdsep, self.deflectionprimemeridian, self.deflectionprimevertical
        # If no subgrids contain coordinates, then raise ValueError
        else:
            raise ValueError("The coordinates supplied are outside the extents of the grid.")


class SubGrid(object):

    def __init__(self):
        self.subName = ''
        self.parent = ''
        self.created = ''
        self.updated = ''
        self.sLat = 0
        self.nLat = 0
        self.eLong = 0
        self.wLong = 0
        self.latInc = 0
        self.longInc = 0
        self.gsCount = 0
        self.Nval = 0
        self.Xi = 0
        self.Eta = 0
        self.contains = False

    def findsubgrid(self, f, lat, lon, interpolation_method='bicubic'):
        """
        Function to pull out metadata of a subgrid in the AUSGeoid2020 NTv2 file

        :param f: variable for open file NTv2 being read as binary
        :param lat: latitude of the point of interest (seconds)
        :param lon: longitude of the point of interest (negative seconds)
        :param interpolation_method: interpolation method as specified in ntv2reader() -
        'bilinear' or 'bicubic' ('bicubic' by default)

        :return: subgrid metadata in form of a dictionary or results from bilinear() or bicubic() (depending on
        interpolation method specified) if point lies within subgrid.
        """
        # SUB_NAME
        f.seek(8, 1)

        byte = f.read(8)
        self.subName = byte.decode('utf').strip('\x00').strip()

        # PARENT
        f.seek(8, 1)

        byte = f.read(8)
        self.parent = byte.decode('utf').strip('\x00').strip()

        # CREATED
        f.seek(8, 1)

        byte = f.read(8)
        self.created = dt.strptime(byte.decode('utf').strip('\x00').strip(), '%d%m%Y').strftime('%d/%m/%Y')

        # UPDATED
        f.seek(8, 1)

        byte = f.read(8)
        self.updated = dt.strptime(byte.decode('utf').strip('\x00').strip(), '%d%m%Y').strftime('%d/%m/%Y')

        # S_LAT
        f.seek(8, 1)

        byte = f.read(8)
        self.sLat = round(struct.unpack('d', byte)[0], 3)

        # N_LAT
        f.seek(8, 1)

        byte = f.read(8)
        self.nLat = round(struct.unpack('d', byte)[0], 3)

        # E_LONG
        f.seek(8, 1)

        byte = f.read(8)
        self.eLong = round(struct.unpack('d', byte)[0], 3)

        # W_LONG
        f.seek(8, 1)

        byte = f.read(8)
        self.wLong = round(struct.unpack('d', byte)[0], 3)

        # LAT_INC
        f.seek(8, 1)

        byte = f.read(8)
        self.latInc = round(struct.unpack('d', byte)[0], 6)

        # LONG_INC
        f.seek(8, 1)

        byte = f.read(8)
        self.longInc = round(struct.unpack('d', byte)[0], 6)

        # GS_COUNT
        f.seek(8, 1)

        byte = f.read(4)
        self.gsCount = int.from_bytes(byte, byteorder='little')

        f.seek(4, 1)

        # Check if coordinates fall within subgrid. If yes, return self.contains = True
        if self.sLat <= lat < self.nLat and self.eLong <= lon < self.wLong:
            self.contains = True

        # If self.contains is True, determine number of columns in grid, and row and column of node to bottom right of
        # point of interest, then call relevant interpolation method function
        if self.contains is True:
            # Determine number of columns
            numcols = 1 + int((self.wLong - self.eLong) / self.longInc)

            # Determine row and col numbers of node below right of point
            row = int((lat - self.sLat) / self.latInc)
            col = int((lon - self.eLong) / self.longInc)

            # If interpolation_method == 'bilinear', call bilinear function
            if interpolation_method == 'bilinear':
                return self.bilinear(f, lat, lon, numcols, row, col)
            # If interpolation_method == 'bicubic', call bicubic function
            elif interpolation_method == 'bicubic':
                return self.bicubic(f, lat, lon, numcols, row, col)

        # If point is not in subgrid, skip to end of subgrid and return sub grid metadata in form of dictionary
        else:
            f.seek(16 * self.gsCount, 1)

            return {'CONTAINS': self.contains, 'SUB_NAME': self.subName, 'PARENT': self.parent, 'CREATED': self.created,
                    'UPDATED': self.updated, 'S_LAT': self.sLat, 'N_LAT': self.nLat, 'E_LONG': self.eLong,
                    'W_LONG': self.wLong, 'LAT_INC': self.latInc, 'LONG_INC': self.longInc, 'GS_COUNT': self.gsCount,
                    'ELLIPSOID_TO_AHD_SEPARATION': np.nan, 'DEFLECTION_PRIME_MERIDIAN': np.nan,
                    'DEFLECTION_PRIME_VERTICAL': np.nan}

    def bilinear(self, f, lat, lon, numcols, row, col):

        """
        Function to perform bilinear interpolatoin of the Ellipsoid to AHD Separation, Deflection in Prime Meridian and
        Deflection in Prime Vertical of a point of interest.

        :param f: variable for open file NTv2 being read as binary
        :param lat: latitude of the point of interest (seconds)
        :param lon: longitude of the point of interest (negative seconds)
        :param numcols: number of columns in grid
        :param row: row number of point to the bottom right of the point of interest
        :param col: column number of point to the bottom right of the point of interest

        :return: dictionary of subgrid metadata and values for Ellipsoid to AHD Separation, Deflection in Prime Meridian
        and Deflection in Prime Vertical of point of interest found via bilinear interpolation.
        """

        #   |     |
        # --o-----o--
        #   |4    |3
        #   | *P  |
        # --o-----o--
        #   |2    |1
        #
        # o - Node position.

        # Determine position in the file of the four surrounding nodes
        pos1 = row * numcols + col
        pos2 = pos1 + 1
        pos3 = pos1 + numcols
        pos4 = pos3 + 1

        # Navigate to start of posA node
        f.seek(16 * pos1, 1)

        # Read in values for nodes A and B
        (pos1nval, pos1xi, pos1eta) = self.read_node(f)
        (pos2nval, pos2xi, pos2eta) = self.read_node(f)

        # Navigate to beginning of node C
        f.seek(16*(pos3 - pos2 - 1), 1)

        # Read in values for nodes C and D
        (pos3nval, pos3xi, pos3eta) = self.read_node(f)
        (pos4nval, pos4xi, pos4eta) = self.read_node(f)

        # Determine latitude and longitude of node A
        lat1 = self.sLat + row * self.latInc
        long1 = self.eLong + col * self.longInc

        # Determine interpolation scale factors
        x = round((lon - long1) / self.longInc, 6)
        y = round((lat - lat1) / self.latInc, 6)

        # Call bilinear interpolation function to determine N value, Xi and Eta
        # (Ellipsoid to AHD separation, deflection in prime meridian and deflection in prime vertical).
        self.Nval = round(bilinear_interpolation(pos1nval, pos2nval, pos3nval, pos4nval, x, y), 3)
        self.Xi = round(bilinear_interpolation(pos1xi, pos2xi, pos3xi, pos4xi, x, y), 2)
        self.Eta = round(bilinear_interpolation(pos1eta, pos2eta, pos3eta, pos4eta, x, y), 2)

        # Navigate to the end of the subgrid
        f.seek(16 * (self.gsCount - pos4 - 1), 1)

        # Return dictionary of information
        return {'CONTAINS': self.contains, 'SUB_NAME': self.subName, 'PARENT': self.parent, 'CREATED': self.created,
                'UPDATED': self.updated, 'S_LAT': self.sLat, 'N_LAT': self.nLat, 'E_LONG': self.eLong,
                'W_LONG': self.wLong, 'LAT_INC': self.latInc, 'LONG_INC': self.longInc, 'GS_COUNT': self.gsCount,
                'ELLIPSOID_TO_AHD_SEPARATION': self.Nval, 'DEFLECTION_PRIME_MERIDIAN': self.Xi,
                'DEFLECTION_PRIME_VERTICAL': self.Eta}

    def bicubic(self, f, lat, lon, numcols, row, col):
        """
        Function to perform bicubic interpolatoin of the Ellipsoid to AHD Separation, Deflection in Prime Meridian and
        Deflection in Prime Vertical of a point of interest.

        :param f: variable for open file NTv2 being read as binary
        :param lat: latitude of the point of interest (seconds)
        :param lon: longitude of the point of interest (negative seconds)
        :param numcols: number of columns in grid
        :param row: row number of point to the bottom right of the point of interest
        :param col: column number of point to the bottom right of the point of interest

        :return: dictionary of subgrid metadata and values for Ellipsoid to AHD Separation, Deflection in Prime Meridian
        and Deflection in Prime Vertical of point of interest found via bicubic interpolation.

        """

        #   |     |     |     |
        # --o-----o-----o-----o--
        #   |11   |12   |13   |14
        #   |     |     |     |
        # --o-----o-----o-----o--
        #   |10   |3    |4    |15
        #   |     | *P  |     |
        # --o-----o-----o-----o--
        #   |9    |2    |1    |16
        #   |     |     |     |
        # --o-----o-----o-----o--
        #   |8    |7    |6    |5
        #
        # o - Node position.

        # Determine position in the file of the sixteen surrounding nodes
        pos1 = row * numcols + col
        pos2 = pos1 + 1
        pos3 = pos2 + numcols
        pos4 = pos3 - 1
        pos5 = pos4 - 2 * numcols - 1
        pos6 = pos5 + 1
        pos7 = pos6 + 1
        pos8 = pos7 + 1
        pos9 = pos8 + numcols
        pos10 = pos9 + numcols
        pos11 = pos10 + numcols
        pos12 = pos11 - 1
        pos13 = pos12 - 1
        pos14 = pos13 - 1
        pos15 = pos14 - numcols
        pos16 = pos15 - numcols

        # Navigate to start of posA node
        f.seek(16 * pos5, 1)

        # Read in values for nodes 5-8
        (pos5nval, pos5xi, pos5eta) = self.read_node(f)
        (pos6nval, pos6xi, pos6eta) = self.read_node(f)
        (pos7nval, pos7xi, pos7eta) = self.read_node(f)
        (pos8nval, pos8xi, pos8eta) = self.read_node(f)

        # Navigate to start of pos16 node
        f.seek(16 * (pos16 - pos8 - 1), 1)

        # Read in values for nodes 16, 1, 2, and 9
        (pos16nval, pos16xi, pos16eta) = self.read_node(f)
        (pos1nval, pos1xi, pos1eta) = self.read_node(f)
        (pos2nval, pos2xi, pos2eta) = self.read_node(f)
        (pos9nval, pos9xi, pos9eta) = self.read_node(f)

        # Navigate to start of pos15 node
        f.seek(16 * (pos15 - pos9 - 1), 1)

        # Read in values for nodes 15, 3, 4 and 10
        (pos15nval, pos15xi, pos15eta) = self.read_node(f)
        (pos4nval, pos4xi, pos4eta) = self.read_node(f)
        (pos3nval, pos3xi, pos3eta) = self.read_node(f)
        (pos10nval, pos10xi, pos10eta) = self.read_node(f)

        # Navigate to start of pos14 node
        f.seek(16 * (pos14 - pos10 - 1), 1)

        # Read in values for nodes 11, 12, 13 and 14
        (pos14nval, pos14xi, pos14eta) = self.read_node(f)
        (pos13nval, pos13xi, pos13eta) = self.read_node(f)
        (pos12nval, pos12xi, pos12eta) = self.read_node(f)
        (pos11nval, pos11xi, pos11eta) = self.read_node(f)

        # Determine latitude and longitude of node A
        lat1 = self.sLat + row * self.latInc
        long1 = self.eLong + col * self.longInc

        # Determine interpolation scale factors
        x = round((lon - long1) / self.longInc, 6)
        y = round((lat - lat1) / self.latInc, 6)

        # Call bicubic_interpolation to determine nVal, Xi and Eta
        # (Ellipsoid to AHD separation, deflection in prime meridian and deflection in prime vertical).
        self.Nval = round(bicubic_interpolation(pos1nval, pos2nval, pos3nval, pos4nval, pos5nval, pos6nval, pos7nval,
                                                pos8nval, pos9nval, pos10nval, pos11nval, pos12nval, pos13nval,
                                                pos14nval, pos15nval, pos16nval, x, y), 3)
        self.Xi = round(bicubic_interpolation(pos1xi, pos2xi, pos3xi, pos4xi, pos5xi, pos6xi, pos7xi, pos8xi, pos9xi,
                                              pos10xi, pos11xi, pos12xi, pos13xi, pos14xi, pos15xi, pos16xi, x, y), 2)
        self.Eta = round(bicubic_interpolation(pos1eta, pos2eta, pos3eta, pos4eta, pos5eta, pos6eta, pos7eta, pos8eta,
                                               pos9eta, pos10eta, pos11eta, pos12eta, pos13eta, pos14eta, pos15eta,
                                               pos16eta, x, y), 2)

        # Navigate to the end of the subgrid
        f.seek(16 * (self.gsCount - pos11 - 1), 1)

        # Return dictionary of information
        return {'CONTAINS': self.contains, 'SUB_NAME': self.subName, 'PARENT': self.parent, 'CREATED': self.created,
                'UPDATED': self.updated, 'S_LAT': self.sLat, 'N_LAT': self.nLat, 'E_LONG': self.eLong,
                'W_LONG': self.wLong, 'LAT_INC': self.latInc, 'LONG_INC': self.longInc, 'GS_COUNT': self.gsCount,
                'ELLIPSOID_TO_AHD_SEPARATION': self.Nval, 'DEFLECTION_PRIME_MERIDIAN': self.Xi,
                'DEFLECTION_PRIME_VERTICAL': self.Eta}

    # Define function to read in node values
    @staticmethod
    def read_node(f):
        """
        Function to read in values of nodes

        :param f:  variable for open file NTv2 being read as binary
        :return: Ellipsoid to AHD Separation, Deflection in Prime Meridian and Deflection in Prime Vertical for node.

        """
        # Read ellipsoid to AHD separation value (N)
        byte = f.read(4)
        nval = round(struct.unpack('f', byte)[0], 3)
        # Read deflection in prime meridian value (Xi)
        byte = f.read(4)
        xi = round(struct.unpack('f', byte)[0], 3)
        # Read deflection in prime vertical value (Eta)
        byte = f.read(4)
        eta = round(struct.unpack('f', byte)[0], 3)
        # Skip to beginning of next node
        f.seek(4, 1)
        # Return values
        return nval, xi, eta


# Define function for bilinear interpolation
def bilinear_interpolation(n1, n2, n3, n4, x, y):
    """
    Bilinear interpolation of value for point of interest (P).

    :param n1: value at node 1
    :param n2: value at node 2
    :param n3: value at node 3
    :param n4: value at node 4
    :param x: interpolation scale factor for x axis
    :param y: interpolation scale factor for y axis

    :return: value at node P

    """

    a0 = n1
    a1 = round(n2 - n1, 3)
    a2 = round(n3 - n1, 3)
    a3 = round(n1 + n4 - n2 - n3, 3)
    p = a0 + (a1 * x) + (a2 * y) + (a3 * x * y)
    return p


# Define function for performing bicubic interpolation
def bicubic_interpolation(n1, n2, n3, n4, n5, n6, n7, n8, n9, n10, n11, n12, n13, n14, n15, n16, x, y):
    """
    Bicubic interpolation of value for point of interest (P).

    :param n1: value at node 1
    :param n2: value at node 2
    :param n3: value at node 3
    :param n4: value at node 4
    :param n5: value at node 5
    :param n6: value at node 6
    :param n7: value at node 7
    :param n8: value at node 8
    :param n9: value at node 9
    :param n10: value at node 10
    :param n11: value at node 11
    :param n12: value at node 12
    :param n13: value at node 13
    :param n14: value at node 14
    :param n15: value at node 16
    :param n16: value at node 17
    :param x: interpolation scale factor for x axis
    :param y: interpolation scale factor for y axis

    :return: value at node P
    """

    # Define the inverse of the coefficient matrix
    cinv = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                     [-3, 0, 0, 3, 0, 0, 0, 0, -2, 0, 0, -1, 0, 0, 0, 0],
                     [2, 0, 0, -2, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
                     [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                     [0, 0, 0, 0, -3, 0, 0, 3, 0, 0, 0, 0, -2, 0, 0, -1],
                     [0, 0, 0, 0, 2, 0, 0, -2, 0, 0, 0, 0, 1, 0, 0, 1],
                     [-3, 3, 0, 0, -2, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, -3, 3, 0, 0, -2, -1, 0, 0],
                     [9, -9, 9, -9, 6, 3, -3, -6, 6, -6, -3, 3, 4, 2, 1, 2],
                     [-6, 6, -6, 6, -4, -2, 2, 4, -3, 3, 3, -3, -2, -1, -1, -2],
                     [2, -2, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 2, -2, 0, 0, 1, 1, 0, 0],
                     [-6, 6, -6, 6, -3, -3, 3, 3, -4, 4, 2, -2, -2, -2, -1, -1],
                     [4, -4, 4, -4, 2, 2, -2, -2, 2, -2, -2, 2, 1, 1, 1, 1]])

    # Define x parameters
    # Function values
    x1 = n1
    x2 = n2
    x3 = n3
    x4 = n4
    # X Derivatives
    x5 = round((n2 - n16) / 2, 4)
    x6 = round((n9 - n1) / 2, 4)
    x7 = round((n10 - n4) / 2, 4)
    x8 = round((n3 - n15) / 2, 4)
    # Y Derivatives
    x9 = round((n4 - n6) / 2, 4)
    x10 = round((n3 - n7) / 2, 4)
    x11 = round((n12 - n2) / 2, 4)
    x12 = round((n13 - n1) / 2, 4)
    # Cross Derivatives (XY)
    x13 = round((n3 - n7 - n15 + n5) / 4, 4)
    x14 = round((n10 - n8 - n4 + n6) / 4, 4)
    x15 = round((n11 - n9 - n13 + n1) / 4, 4)
    x16 = round((n12 - n2 - n14 + n16) / 4, 4)

    # Create array from x parameters
    xarr = np.array([x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15, x16])

    # Multiply the inverse of the coefficient matrix by the array of x values to give array of alpha values
    alpha = np.matmul(cinv, xarr)

    # Calculate value at the point of interest
    n_p = 0
    for i in range(0, 4):
        for j in range(0, 4):
            n_p = n_p + alpha[i * 4 + j] * x ** i * y ** j

    # Return the value
    return n_p


# TEST OF SCRIPT
# Specify AUSGeoid2020 binary NTv2 file location
ntv2file = "C://Git/Python/NTv2.git/AUSGeoid2020_20180201.gsb"
# Assign class to variable
grids = NTv2ReaderBinary()
# Call ntv2reader to determine values at specific point
values = grids.ntv2reader(ntv2file, -26.948643, 145.6548721, 'bilinear')
# Print values
print(values)
